/*
 * @(#)DBSession.java	2010-4-21
 *
 * Copyright (c) 2010 by gerald. All Rights Reserved.
 */

package org.javaclub.jorm;

import java.io.Serializable;
import java.sql.Connection;
import java.util.List;
import java.util.Map;

import org.javaclub.jorm.jdbc.BeanPersister;
import org.javaclub.jorm.jdbc.JdbcException;
import org.javaclub.jorm.jdbc.batch.JdbcBatcher;
import org.javaclub.jorm.jdbc.callable.ProcedureCaller;
import org.javaclub.jorm.jdbc.sql.Dialect;
import org.javaclub.jorm.jdbc.sql.SqlParams;

/**
 * A jdbc session deal with database.
 *
 * @author <a href="mailto:gerald.chen@qq.com">Gerald Chen</a>
 * @version $Id: Session.java 545 2011-10-09 07:53:09Z gerald.chen.hz $
 */
@SuppressWarnings("unchecked")
public interface Session {
	
	/**
	 * Get the underlying database dialect.
	 *
	 * @return the underlying database dialect.
	 */
	public Dialect getDialect();
	
	/**
	 * Get a jdbc connection from this Session.
	 *
	 * @return JDBC Connection
	 */
	public Connection getConnection();
	
	/**
	 * Require a new JDBC Connection.
	 *
	 * @param isnew is this required session is fully new ?
	 * @return JDBC Connection
	 * @throws JdbcException
	 */
	public Connection getConnection(boolean isnew) throws JdbcException;
	
	/**
	 * Start transaction.
	 *
	 * @throws JdbcException
	 */
	public void beginTransaction() throws JdbcException;
	
    /**
     * Commit the transaction.
     *
     * @throws JdbcException
     */
    public void commit() throws JdbcException;
    
    /**
     * If operation failed, do rollback.
     *
     * @throws JdbcException
     */
    public void rollback() throws JdbcException;
    
    /**
     * Ending transaction.
     *
     * @throws JdbcException
     */
    public void endTransaction() throws JdbcException;
    
    /**
     * Close the JdbcSession
     *
     * @throws JdbcException
     */
    public void close() throws JdbcException;
    
    /**
     * Present the session is auto-closed type or not.
     *
     * @return true if this session is auto-closed type, otherwise false
     * @since 1.0.7
     */
    public boolean isAutoClose();
    
    /**
     * This method is used to call a sql stored procedure.
     *
     * @param caller a abstract procedure caller param
     * @return the invoked result of a sql stored procedure
     * @throws JdbcException
     */
    public <T> T call(ProcedureCaller caller) throws JdbcException;
    
    /**
     * Executes a stored procedure that without a returning value.
     *
     * @param procedure the stored procedure
     * @param params input parameters
     * @throws JdbcException
     */
    public void call(final String procedure, final Object... params) throws JdbcException;
    
    public void flush() throws JdbcException;
    
    /**
     * Gets the default connection's database name.
     *
     * @return the default connection's database name
     * @throws JormException
     */
    public String getDefaultDatabase() throws JormException;
    
    /**
     * Gets the specified connection's database name.
     *
     * @param provider a specified connection provider name
     * @return database name
     * @throws JormException
     */
    public String getDatabase(String provider) throws JormException;
    
    /**
     * Creates a JDBC Batch operator.
     *
     * @return JDBC Batch operator
     * @throws JdbcException
     */
    public JdbcBatcher createBatcher() throws JdbcException;
    
    /**
     * Gets the holding entity persiter of session
     *
     * @return a entity persiter
     */
    public BeanPersister getPersister();
	
	/**
	 * If the JdbcSession is closed.
	 *
	 * @return true if JdbcSession is closed, othwise false.
	 */
	public boolean isClosed();
	
	/**
	 * Reads a list of all objects in the database mapped to the given object class.
	 *
	 * @param objectClass entity class type
	 * @return the collection result
	 * @throws JdbcException
	 */
	public <T> List<T> all(Class<T> objectClass) throws JdbcException;
	
	/**
	 * Persist the given transient instance, first assigning a generated identifier. (Or
	 * using the current value of the identifier property if the <tt>assigned</tt>
	 * generator is used.)
	 *
	 * @param object a transient instance of a persistent class
	 * @return the generated identifier
	 * @throws JdbcException
	 */
	public Serializable save(Object object) throws JdbcException;

	/**
	 * Update the persistent instance with the identifier of the given detached
	 * instance. 
	 *
	 * @param object a detached instance containing updated state
	 * @throws JdbcException
	 */
	public <T> void update(T object) throws JdbcException;
	
	/**
	 * Either {@link #save(Object)} or {@link #update(Object)} the given
	 * instance, depending upon resolution of the unsaved-value checks (see the
	 * manual for discussion of unsaved-value checking).
	 * <p/>
	 *
	 * @see Session#save(java.lang.Object)
	 * @see Session#update(Object object)
	 * @param object a transient or detached instance containing new or updated state
	 * @throws JdbcException
	 */
	public void saveOrUpdate(Object object) throws JdbcException;
	
	/**
	 * Clean all the data records of a table.
	 *
	 * @param objectClass the mapping entity of the target table
	 * @return influenced record count
	 * @throws JdbcException
	 */
	public <T> int clean(Class<T> objectClass) throws JdbcException;
	
	/**
	 * Clean all the data records of a table.
	 *
	 * @param tbname table name
	 * @return influenced record count
	 * @throws JdbcException
	 */
	public int clean(String tbname) throws JdbcException;
	
	/**
	 * Tests if the specfied object exsits in database.
	 *
	 * @param object the object to test
	 * @return true if the object exsits in database, false otherwise.
	 * @throws JdbcException
	 */
	public <T> boolean has(T object) throws JdbcException;
	
	/**
	 * Remove a persistent instance from the database. The argument may be
	 * an instance associated with the receiving <tt>Session</tt> or a transient
	 * instance with an identifier associated with existing persistent state.
	 *
	 * @param object the instance to be removed
	 * @throws JdbcException
	 */
	public <T> void delete(T object) throws JdbcException;
	
	/**
	 * Execute delete operation with sql only.
	 *
	 * @param sql sql
	 * @return influenced record count
	 * @throws JdbcException
	 */
	public int delete(String sql) throws JdbcException;
	
	/**
	 * Delete all objects returned by the query. Return the number of objects deleted.
	 * 
	 */
	public int delete(SqlParams params) throws JdbcException;
	
	/**
	 * Delete objects by conditions.
	 *
	 * @param <T>
	 * @param objectClass the class to delete.
	 * @param whereFragment conditions
	 * @return the successfully deleting results count.
	 * @throws JdbcException
	 */
	public <T> int delete(Class<T> objectClass, String whereFragment) throws JdbcException;
	
	/**
	 * Query object by primary key.
	 *
	 * @param <T>
	 * @param objectClass the class to lookup
	 * @param id primary key
	 * @return query result
	 * @throws JdbcException
	 */
	public <T> T read(Class<T> objectClass, Serializable id) throws JdbcException;
	
	/**
	 * 根据复合主键，获取一个对象，该对象必须声明 '@PK'.
	 * <p>
	 * 并且：给定的参数顺序 必须同 '@PK' 中声明的顺序一致，否则会产生不可预知的错误。
	 * </p>
	 *
	 * @param objectClass
	 * @param pks  复合主键
	 * @return  查找结果
	 * @throws JdbcException
	 */
	public <T> T read(Class<T> objectClass, Object... pks) throws JdbcException;
	
	/**
	 * Loads the frist record object.
	 *
	 * @param sqlParams query parameters
	 * @return domain object
	 * @throws JdbcException
	 */
	public <T> T loadFirst(SqlParams<T> sqlParams) throws JdbcException;
	
	/**
	 * Loading the first Java entity that satisfy query conditions
	 *
	 * @param objectClass the entity class type
	 * @param sql query sql statement
	 * @param params bind parameters
	 * @return the first Java entity
	 * @throws JdbcException
	 */
	public <T> T loadFirst(Class<T> objectClass, String sql, Object... params) throws JdbcException;
	
	/**
	 * Load entities that satisfy specified conditions.
	 *
	 * @param params SqlParams
	 * @return collection result
	 * @throws JdbcException
	 */
	public <T> List<T> loadBeans(SqlParams<T> params) throws JdbcException;

	/**
	 * Find Objects by condition.
	 *
	 * @param objectClass the entity class type
	 * @param params conditions
	 * @return query objects
	 * @throws JdbcException
	 */
	public <T> List<T> list(SqlParams<T> params) throws JdbcException;
	
	/**
	 * Find entities that satisfy specified conditions.
	 *
	 * @param objectClass the entity class type
	 * @param sql query sql statement
	 * @param params bind parameters
	 * @return collection result
	 * @throws JdbcException
	 */
	public <T> List<T> list(Class<T> objectClass, String sql, Object... params) throws JdbcException;
	
	// --- JDBC Operation ---
	/**
	 * Executes the given INSERT, UPDATE, or DELETE SQL statement without any replacement parameters.
	 * 
	 * @param sql The SQL statement to execute.
	 * @return The number of rows updated.
	 */
	public int executeUpdate(String sql) throws JdbcException;

	/**
	 * 执行带占位符的更新语句
	 * 
	 * @param sql 要执行的SQL语句
	 * @param params 占位符所对应的值Object
	 * @return 返回影响记录条数 , 默认为0
	 */
	public int executeUpdate(String sql, Object[] params) throws JdbcException;

	/**
	 * Executes sql with some bind parameters.
	 * 
	 * @param sqlParams bind SqlParams
	 * @return the count of the influenced rows
	 */
	public int executeUpdate(SqlParams sqlParams) throws JdbcException;

	/**
	 * 批量更新操作,多条SQL语句,失败则全部回滚
	 * 
	 * @param sqls SQL语句数组
	 * @return 返回受影响的记录数,int[i]表示sqls[i]的影响记录数
	 */
	public int[] executeBatch(String[] sqls) throws JdbcException;
	
	/**
	 * 批量更新操作,同条SQL语句,单一不同参数,失败则全部回滚
	 * 
	 * @param sql SQL语句
	 * @param params 不同的参数值
	 * @param perBatchSize 批量大小,0直接返回空列表
	 * @return 返回影响记录数List<int[]>,List[i]表示第i批影响的记录数int[]
	 */
	public List<int[]> executeBatch(String sql, Object[] params, int perBatchSize) throws JdbcException;

	/**
	 * 批量更新操作,同条SQL语句,多个不同参数,失败则全部回滚
	 * 
	 * @param sql SQL语句
	 * @param params 占位符参数
	 * @param perBatchSize 批量大小,0直接返回空列表
	 * @return 返回影响记录数List<int[]>,List[i]表示第i批影响的记录数int[]
	 */
	public List<int[]> executeBatch(String sql, Object[][] params, int perBatchSize) throws JdbcException;

	/**
	 * Batch insert operation.
	 * 
	 * @param sql 带占位符的sql语句
	 * @param datalist 要插入的数据列表
	 * @return 是否操作成功
	 */
	public boolean batchInsert(String sql, List<Object[]> datalist) throws JdbcException;
	
	/**
	 * Batch insert operation.
	 *
	 * @param sql 带占位符的sql语句
	 * @param datalist 要插入的数据列表
	 * @param perBatchSize 每perBatchSize大小作为一个单元
	 * @return 是否操作成功
	 */
	public boolean batchInsert(String sql, List<Object[]> datalist, int perBatchSize) throws JdbcException;
	
	/**
	 * 删除记录
	 *
	 * @param tableName 表名
	 * @param conditions 条件
	 * @return 影响的记录数量
	 */
	public int delete(String tableName, Map<String, Object> conditions) throws JdbcException;

	/**
	 * 动态更新某些列的值
	 * 
	 * @param tableName 表名
	 * @param targetColumnNamesAndValues 要更新的列名和对应的值
	 * @param paramColumnAndValues 条件（包括列名和对应值）
	 * @return 影响的记录数量
	 */
	public int dynamicUpdate(String tableName, Map<String, Object> targetColumnNamesAndValues,
			Map<String, Object> paramColumnAndValues) throws JdbcException;

	/**
	 * 往表里插入一条新记录，只给特定的列赋值
	 * 
	 * @param tableName 数据表名
	 * @param columnNames 列名
	 * @param columnValues 对应的列值
	 * @return 影响的记录数量
	 */
	public int dynamicInsert(String tableName, String[] columnNames, Object[] columnValues) throws JdbcException;
	
	/**
	 * Get total records count by the specified jdbc sql.
	 *
	 * @param sql jdbc query sql statement
	 * @param params bind parameters
	 * @return records count that satisfy condition
	 * @throws JdbcException
	 */
	public long count(String sql, Object... params) throws JdbcException;

	/**
	 * Get total records count by the specified wraped jdbc sql.
	 * 
	 * @param sqlParams the wraped jdbc sql
	 * @return total records count 
	 */
	public long count(SqlParams sqlParams) throws JdbcException;
	
	public long count(Class<?> objectClass) throws JdbcException;
	
	/**
	 * Test a table whether contains the specified record or not.
	 *
	 * @param tableName table name
	 * @param conditions query condition with the form of column and value
	 * @return <code>true</code> if the specified table has the record that satisfy conditions, 
	 *         otherwise <code>false</code>.
	 */
	public boolean has(String tableName, Map<String, Object> conditions) throws JdbcException;
	
	/**
	 * Test a table whether contains the specified record or not.
	 *
	 * @param tableName table name
	 * @param conditions query condition
	 * @param inConditions query condition of IN(...)
	 * @return <code>true</code> if the specified table has the record that satisfy conditions, 
	 *         otherwise <code>false</code>.
	 */
	public boolean has(String tableName, Map<String, Object> conditions, Map<String, Object[]> inConditions) throws JdbcException;
	
	/**
	 * Test current database whether contains the specified table or not.
	 *
	 * @param tbname table name
	 * @return <code>true</code> if current database contains the specified table, otherwise <code>false</code>.
	 */
	public boolean existsTable(String tbname) throws JdbcException;
	
	/**
	 * Test the database whether contains the specified table or not.
	 * <p>
	 * for PostgreSQL, if the database is not the current connection's database, this method will not take effect.
	 *
	 * @param database database name
	 * @param tbname table name
	 * @return  <code>true</code> if database contains the specified table with name tbname, otherwise <code>false</code>.
	 */
	public boolean existsTable(String database, String tbname) throws JdbcException;
	
	/**
	 * Used to load a single column or entity, such as following sqls: 
	 * <li> select name from User where ID = 1 
	 * <li> select t.* from t_user t limit 1
	 * 
	 * @param sql query sql statement
	 */
	public <T> T unique(String sql) throws JdbcException;

	/**
	 * Used to load a single column or entity, such as following sqls: 
	 * <li> select name from User where ID = ? 
	 * <li> select t.* from t_user t where t.name = ? limit 1
	 * 
	 * @param sqlParams sql params
	 */
	public <T> T unique(SqlParams sqlParams) throws JdbcException;
	
	public Object column(String sql, String colAlias, Object... params) throws JdbcException;

	public Object column(String colAlias, SqlParams sqlParams) throws JdbcException;
	
	/**
	 * 根据SQL,返回第一行查询结果,Map<field,value>
	 * 
	 * @param sql:SQL语句
	 * @return Map<field,value>:(表信息列表)
	 */
	public Map<String, Object> queryMap(String sql, Object... params) throws JdbcException;

	/**
	 * 根据SQL,返回第一行查询结果,Map<field,value>
	 * 
	 * @param sqlParams:SQL语句
	 * @return Map<field,value>:(表信息列表)
	 */
	public Map<String, Object> queryMap(SqlParams sqlParams) throws JdbcException;
	
	/**
	 * 根据SQL,返回查询结果,List<Map<field,value>>
	 * 
	 * @param sql:SQL语句
	 * @return list:(表信息列表，一条信息用一个Map代表)
	 */
	public List<Map<String, Object>> queryMapList(String sql, Object... params) throws JdbcException;

	/**
	 * 根据SQL,返回查询结果,List<Map<field,value>>
	 * 
	 * @param sqlParams:SQL语句
	 * @return list:(表信息列表，一条信息用一个Map代表)
	 */
	public List<Map<String, Object>> queryMapList(SqlParams sqlParams) throws JdbcException;
	
	/**
	 * 返回第一列与第二列的MAP，以第一列为KEY，第二列为VALUE
	 * 
	 * @param sql 要执行的sql语句,查询字段必须为2个或以上，否则将返回为空的MAP
	 * @return Map<K, V>
	 */
	public <K, V> Map<K, V> queryColumnsAsMap(String sql, Object... params) throws JdbcException;

	/**
	 * 返回第一列与第二列的MAP，以第一列为KEY，第二列为VALUE
	 * 
	 * @param sqlParams JdbcSql,查询字段必须为2个或以上，否则将返回为空的MAP
	 * @return Map<K, V>
	 */
	public <K, V> Map<K, V> queryColumnsAsMap(SqlParams sqlParams) throws JdbcException;
	
}
